-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de) 
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  25.12.2017  mifi  First Version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity top is
   port (
      --
      -- Input clock
      --
      CLOCK_50   : in  std_logic;
      
      --
      -- User LEDs
      --
      LED         : out std_logic_vector(7 downto 0);
      
      --
      -- Keys
      --
      KEY         : in  std_logic_vector(1 downto 0);
      
      --
      -- SPI
      --
      SPI_CLK     : in  std_logic;
      SPI_CS      : in  std_logic;
      SPI_MOSI    : in  std_logic;
      SPI_MISO    : out std_logic;
      
      --
      -- RESET
      --
      RESET_n     : in  std_logic;
      
      --
      -- Timer
      --
      TIMER_INT_n : out std_logic
   );
end entity top;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of top is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component wb_syscon is
      port (
         -- None Wishbone
         clk_i    : in  std_logic := '0';
         reset_i  : in  std_logic := '0';
         clk_16_o : out std_logic;

         -- Wishbone
         clk_o    : out std_logic;
         rst_o    : out std_logic
      );
   end component wb_syscon;


   component heartbeat is
      generic (
         SYS_CLK_HZ : integer;
         PERIOD_MS  : integer
      );
      port (  
         clk_i : in  std_logic := '0';
         led_o : out std_logic
      );
   end component heartbeat;
   
   
   component wb_core is
      generic (
         ADDRESS_WIDTH : integer;
         DATA_WIDTH    : integer
      );
      port ( 
         -- Syscon
         clk_i        : in  std_logic := '0';
         rst_i        : in  std_logic := '0';

         -- SPI bus
         spi_clk_i    : in  std_logic := '0';
         spi_cs_i     : in  std_logic := '1';
         spi_mosi_i   : in  std_logic := '0';
         spi_miso_o   : out std_logic;

         -- GPIO
         gpio_o       : out std_logic_vector(DATA_WIDTH-1 downto 0);
         gpio_i       : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');

         -- Timer
         timer_zero_o : out std_logic;
         timer_int_o  : out std_logic
      );
   end component wb_core;


   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   signal wb_clk    : std_logic := '0';
   signal wb_rst    : std_logic := '0';
   signal clk_16    : std_logic := '0';

   signal reset     : std_logic := '0';
   signal gpio      : std_logic_vector(7 downto 0) := (others => '0');
   signal timer_int : std_logic := '0';
   
begin

   reset <= '1' when ((RESET_n = '0') OR (KEY(1) = '0') OR (KEY(0) = '0')) else '0';

   --
   -- SYSCON
   --
   inst_syscon : wb_syscon 
      port map ( 
         -- None Wishbone
         clk_i    => CLOCK_50,
         reset_i  => reset,
         clk_16_o => clk_16,

         -- Wishbone
         clk_o    => wb_clk,
         rst_o    => wb_rst
      );


   --   
   -- Heartbeat
   --
   inst_heartbeat : heartbeat 
      generic map (
         SYS_CLK_HZ => 16000000,
         PERIOD_MS  => 1000
      )
      port map ( 
         clk_i => clk_16,
         led_o => LED(7)
      );


   --
   -- Wishbone core
   --
   inst_wb_core : wb_core
      generic map (
         ADDRESS_WIDTH => 13,
         DATA_WIDTH    => 8
      )
      port map (
         -- Syscon
         clk_i        => wb_clk,
         rst_i        => wb_rst,
         
         -- SPI bus
         spi_clk_i    => SPI_CLK,
         spi_cs_i     => SPI_CS,
         spi_mosi_i   => SPI_MOSI,
         spi_miso_o   => SPI_MISO,
         
         -- GPIO
         gpio_o       => gpio,
         gpio_i       => gpio,
         
         -- Timer
         timer_zero_o => open,
         timer_int_o  => timer_int
      );


   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   LED(6 downto 0) <= gpio(6 downto 0);
   
   TIMER_INT_n <= NOT timer_int;
         
end architecture syn;

-- *** EOF ***

