-- ***************************************************************************
-- *  Copyright (c) 2015-2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  21.07.2015  mifi  First version.
-- *  26.12.2017  mifi  Added DATA_WIDTH support.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity wb_core is
   generic (
      ADDRESS_WIDTH : integer := 8;
      DATA_WIDTH    : integer := 16
   );
   port (    
      -- Syscon
      clk_i        : in  std_logic := '0';
      rst_i        : in  std_logic := '0';

      -- SPI bus
      spi_clk_i    : in  std_logic := '0';
      spi_cs_i     : in  std_logic := '1';
      spi_mosi_i   : in  std_logic := '0';
      spi_miso_o   : out std_logic;
      
      -- GPIO
      gpio_o       : out std_logic_vector(DATA_WIDTH-1 downto 0);
      gpio_i       : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
      
      -- Timer
      timer_zero_o : out std_logic;
      timer_int_o  : out std_logic
   );
end entity wb_core;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of wb_core is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component spi_wb_master is
      generic (
         ADDRESS_WIDTH : integer
      );
      port (
         -- System  
         clk_i      : in  std_logic := '0';
         rst_i      : in  std_logic := '0';
      
         -- SPI Interface
         spi_clk_i  : in  std_logic := '0';
         spi_cs_i   : in  std_logic := '1';
         spi_mosi_i : in  std_logic := '0';
         spi_miso_o : out std_logic;
      
         -- Wishbone Master
         wbm_stb_o  : out std_logic;
         wbm_cyc_o  : out std_logic;
         wbm_we_o   : out std_logic;
         wbm_ack_i  : in  std_logic := '0';

         wbm_adr_o  : out std_logic_vector(ADDRESS_WIDTH-1 downto 0);  
         
         wbm_dat_o  : out std_logic_vector(7 downto 0);  
         wbm_dat_i  : in  std_logic_vector(7 downto 0) := (others => '0')
      );
   end component spi_wb_master;
   
   
   component wb_intercon is
      generic (
         ADDRESS_WIDTH : integer;
         DATA_WIDTH    : integer
      );
      port (  
         -- Syscon
         clk_i      : in  std_logic := '0';
         rst_i      : in  std_logic := '0';
      
         -- Wishbone Master
         wbm_stb_i  : in  std_logic := '0';
         wbm_cyc_i  : in  std_logic := '0';
         wbm_we_i   : in  std_logic := '0';
         wbm_ack_o  : out std_logic;
         wbm_adr_i  : in  std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
         wbm_dat_i  : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         wbm_dat_o  : out std_logic_vector(DATA_WIDTH-1 downto 0);
         
         -- Wishbone Slave x
         wbs_cyc_o  : out std_logic;
         wbs_we_o   : out std_logic; 
         wbs_dat_o  : out std_logic_vector(DATA_WIDTH-1 downto 0);           
         wbs_adr_o  : out std_logic_vector(7 downto 0);  
      
         -- Wishbone Slave 1
         wbs1_stb_o : out std_logic;
         wbs1_ack_i : in  std_logic := '0';
         wbs1_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         
         -- Wishbone Slave 2
         wbs2_stb_o : out std_logic;
         wbs2_ack_i : in  std_logic := '0';
         wbs2_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         
         -- Wishbone Slave 3
         wbs3_stb_o : out std_logic;
         wbs3_ack_i : in  std_logic := '0';
         wbs3_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0')
      );
   end component wb_intercon;
   
   
   component wb_test is
      generic (
         DATA_WIDTH : integer
      );
      port (  
         -- Syscon
         clk_i    : in  std_logic := '0';
         rst_i    : in  std_logic := '0';
      
         -- Wishbone
         wb_stb_i : in  std_logic := '0';
         wb_cyc_i : in  std_logic := '0';      -- Not used here
         wb_we_i  : in  std_logic := '0';
         wb_ack_o : out std_logic;
         wb_adr_i : in  std_logic_vector(2 downto 0)  := (others => '0');
         wb_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         wb_dat_o : out std_logic_vector(DATA_WIDTH-1 downto 0)
      );
   end component wb_test;
   

   component wb_gpio is
      generic (
         DATA_WIDTH : integer
      );
      port (  
         -- Syscon
         clk_i    : in  std_logic := '0';
         rst_i    : in  std_logic := '0';
      
         -- Wishbone
         wb_stb_i : in  std_logic := '0';
         wb_cyc_i : in  std_logic := '0';      -- Not used here
         wb_we_i  : in  std_logic := '0';
         wb_ack_o : out std_logic;
         wb_adr_i : in  std_logic_vector(2 downto 0)  := (others => '0');
         wb_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         wb_dat_o : out std_logic_vector(DATA_WIDTH-1 downto 0);  
            
         -- None Wishbone
         port_i   : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         port_o   : out std_logic_vector(DATA_WIDTH-1 downto 0)
      );
   end component wb_gpio;
   
   
   component wb_timer is
      generic (
         DATA_WIDTH : integer := 8
      );
      port (  
         -- Syscon
         clk_i    : in  std_logic := '0';
         rst_i    : in  std_logic := '0';
      
         -- Wishbone
         wb_stb_i : in  std_logic := '0';
         wb_cyc_i : in  std_logic := '0';      -- Not used here
         wb_we_i  : in  std_logic := '0';
         wb_ack_o : out std_logic;
         wb_adr_i : in  std_logic_vector(2 downto 0)            := (others => '0');
         wb_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
         wb_dat_o : out std_logic_vector(DATA_WIDTH-1 downto 0);  
            
         -- None Avalon
         zero_o   : out std_logic;
         int_o    : out std_logic
      );
   end component wb_timer;
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   signal wbm_stb        : std_logic := '0';
   signal wbm_cyc        : std_logic := '0';
   signal wbm_we         : std_logic := '0';
   signal wbm_ack        : std_logic := '0';
   signal wbm_adr        : std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
   signal wbm_writedata  : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal wbm_readdata   : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal wbs_cyc        : std_logic := '0';
   signal wbs_we         : std_logic := '0';
   signal wbs_writedata  : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal wbs_adr        : std_logic_vector(7 downto 0)  := (others => '0');
   

   signal test_stb       : std_logic := '0';
   signal test_ack       : std_logic := '0';
   signal test_readdata  : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal gpio_stb       : std_logic := '0';
   signal gpio_ack       : std_logic := '0';
   signal gpio_readdata  : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal gpio_out       : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal timer_stb      : std_logic := '0';
   signal timer_ack      : std_logic := '0';
   signal timer_readdata : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal timer_zero     : std_logic := '0';
   signal timer_int      : std_logic := '0';
     
begin

   -- Wishbone Master
   inst_master : spi_wb_master       
      generic map (
         ADDRESS_WIDTH => ADDRESS_WIDTH
      )
      port map (  
         -- Syscon
         clk_i      => clk_i,
         rst_i      => rst_i,
         
         -- None Wishbone
         spi_clk_i  => spi_clk_i,
         spi_cs_i   => spi_cs_i,
         spi_mosi_i => spi_mosi_i,
         spi_miso_o => spi_miso_o,
      
         -- Wishbone Master
         wbm_stb_o  => wbm_stb,
         wbm_cyc_o  => wbm_cyc,
         wbm_we_o   => wbm_we,
         wbm_ack_i  => wbm_ack,
         wbm_adr_o  => wbm_adr,
         wbm_dat_o  => wbm_writedata,
         wbm_dat_i  => wbm_readdata 
      );
      
   -- Wishbone Intercon
   inst_intercon : wb_intercon      
      generic map (
         ADDRESS_WIDTH => ADDRESS_WIDTH,
         DATA_WIDTH    => DATA_WIDTH
      )
      port map (  
         -- Syscon
         clk_i       => clk_i,
         rst_i       => rst_i,
      
         -- Wishbone Master
         wbm_stb_i   => wbm_stb,
         wbm_cyc_i   => wbm_cyc,
         wbm_we_i    => wbm_we,
         wbm_ack_o   => wbm_ack,
         wbm_adr_i   => wbm_adr,
         wbm_dat_i   => wbm_writedata,
         wbm_dat_o   => wbm_readdata,
         
         -- Wishbone Slave x
         wbs_cyc_o   => wbs_cyc,
         wbs_we_o    => wbs_we,
         wbs_dat_o   => wbs_writedata,
         wbs_adr_o   => wbs_adr,
      
         -- Wishbone Slave 1
         wbs1_stb_o  => test_stb,
         wbs1_ack_i  => test_ack,
         wbs1_dat_i  => test_readdata,
         
         -- Wishbone Slave 2
         wbs2_stb_o  => gpio_stb,
         wbs2_ack_i  => gpio_ack,
         wbs2_dat_i  => gpio_readdata,
         
         -- Wishbone Slave 3
         wbs3_stb_o  => timer_stb,
         wbs3_ack_i  => timer_ack,
         wbs3_dat_i  => timer_readdata
      );
      
   -- Wishbone TEST
   int_test : wb_test
      generic map (
         DATA_WIDTH => DATA_WIDTH
      )
      port map (  
         -- Syscon
         clk_i    => clk_i,
         rst_i    => rst_i,
      
         -- Wishbone
         wb_stb_i => test_stb,
         wb_cyc_i => wbs_cyc,
         wb_we_i  => wbs_we,
         wb_ack_o => test_ack,
         wb_adr_i => wbs_adr(2 downto 0),
         wb_dat_i => wbs_writedata,
         wb_dat_o => test_readdata
      );
      
   -- Wishbone GPIO
   inst_gpio : wb_gpio
      generic map (
         DATA_WIDTH => DATA_WIDTH
      )
      port map (  
         -- Syscon
         clk_i    => clk_i,
         rst_i    => rst_i,
      
         -- Wishbone
         wb_stb_i => gpio_stb,
         wb_cyc_i => wbs_cyc,
         wb_we_i  => wbs_we,
         wb_ack_o => gpio_ack,
         wb_adr_i => wbs_adr(2 downto 0),
         wb_dat_i => wbs_writedata,
         wb_dat_o => gpio_readdata,
            
         -- None Wishbone
         port_i   => gpio_i, 
         port_o   => gpio_out
      );
      
   -- Wishbone TIMER     
   int_timer : wb_timer
      generic map (
         DATA_WIDTH => DATA_WIDTH
      )
      port map (  
         -- Syscon
         clk_i    => clk_i,
         rst_i    => rst_i,
      
         -- Wishbone
         wb_stb_i => timer_stb,
         wb_cyc_i => wbs_cyc,
         wb_we_i  => wbs_we,
         wb_ack_o => timer_ack,
         wb_adr_i => wbs_adr(2 downto 0),
         wb_dat_i => wbs_writedata,
         wb_dat_o => timer_readdata,
            
         -- None Avalon
         zero_o   => timer_zero,
         int_o    => timer_int
      );
      
               
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   gpio_o       <= gpio_out;
   timer_zero_o <= timer_zero;
   timer_int_o  <= timer_int;
         
end architecture syn;

-- *** EOF ***

