-- ***************************************************************************
-- *  Copyright (c) 2015-2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  20.07.2015  mifi  First version.
-- *  27.12.2017  mifi  Changed for the SPI Wishbone Master example.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity wb_gpio is
   generic (
      DATA_WIDTH : integer := 8
   );
   port (  
      -- Syscon
      clk_i    : in  std_logic := '0';
      rst_i    : in  std_logic := '0';
      
      -- Wishbone
      wb_stb_i : in  std_logic := '0';
      wb_cyc_i : in  std_logic := '0';      -- Not used here
      wb_we_i  : in  std_logic := '0';
      wb_ack_o : out std_logic;
      wb_adr_i : in  std_logic_vector(2 downto 0)            := (others => '0');
      wb_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
      wb_dat_o : out std_logic_vector(DATA_WIDTH-1 downto 0);  
            
      -- None Wishbone
      port_i   : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
      port_o   : out std_logic_vector(DATA_WIDTH-1 downto 0)
   );
end entity wb_gpio;


-- *****************************************************************************
-- *  DEFINE: Architecture                                                     *
-- *****************************************************************************

architecture syn of wb_gpio is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant ADR_DATAOUT : std_logic_vector(2 downto 0) := "000"; -- Read / Write
   constant ADR_DATAIN  : std_logic_vector(2 downto 0) := "001"; -- Read
   constant ADR_SET     : std_logic_vector(2 downto 0) := "010"; -- Write only
   constant ADR_CLEAR   : std_logic_vector(2 downto 0) := "011"; -- Write only
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   signal out_reg    : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal in_reg     : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal readdata   : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal read_mux   : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal data_in_s1 : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');

   signal read_req   : std_logic := '0'; 
   signal read_ack   : std_logic := '0';
   
   signal write_req  : std_logic := '0';
   signal write_ack  : std_logic := '0';
   
begin

   -- Some simplifications for later use
   read_req  <= '1' when ((wb_stb_i = '1') and (wb_we_i = '0')) else '0';
   write_req <= '1' when ((wb_stb_i = '1') and (wb_we_i = '1')) else '0';

   --
   -- Register File
   --
   process (clk_i, rst_i)
   begin
      if (rst_i = '1') then

         out_reg   <= (others => '0');
         write_ack <= '0';
      
      elsif rising_edge(clk_i) then
      
         -- Default
         write_ack <= '0';
      
         if (write_req = '1') then
         
            write_ack <= '1';    -- Signal normal Wishbone bus cycle termination
         
            -- Address decoder
            case wb_adr_i is
            
               when ADR_DATAOUT =>
                  out_reg <= wb_dat_i;
            
               when ADR_SET =>
                  out_reg <= out_reg OR wb_dat_i;

               when ADR_CLEAR =>
                  out_reg <= out_reg AND (NOT wb_dat_i);
                  
               when others =>
                  -- Do nothing
            
            end case;
         end if;
         
      end if; 
   end process;
   
   --
   -- Synchronize the input data
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         data_in_s1 <= port_i;
         in_reg     <= data_in_s1;
      end if;
   end process;
   
   --
   -- Read Multiplexer
   --
   read_mux <= out_reg when (ADR_DATAOUT = wb_adr_i) else 
               in_reg  when (ADR_DATAIN  = wb_adr_i) else (others => '0'); 

   process (clk_i)
   begin
      if rising_edge(clk_i) then
         -- Default
         read_ack <= '0';        
      
         if (read_req = '1') then
            readdata <= read_mux;
            read_ack <= '1';     -- Signal normal Wishbone bus cycle termination
         end if;
      end if;
   end process;      

   
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- Wishbone
   wb_ack_o <= read_ack or write_ack;
   wb_dat_o <= readdata;

   -- None Wishbone
   port_o <= out_reg;
   
end architecture syn;

-- *** EOF ***

